/*************************************************************************************
 Data Visualization - Checkout length of Dewey Categories              
 Tested in Processing 3                                    
 
 Author: Wilson Mui                       
 Adapted from: Rodger (Jieliang) Luo, George Legrady                           
 
 Purpose: See how quickly people return books for various large Dewey categories.   
 
 Usage: Enter categories in textbox to compare various categories alongside eachother.
 
 Notes:
 
 PeasyCam and controlP5 libraries must be installed for this to run.
 
 Processing 3.4 was used to create this. However, if the program fails to run,
 updating to 3.5 should fix it. It failed to run on Ubuntu with 3.4.
 *************************************************************************************/
import controlP5.*; //GUI library
import java.util.*; 
import peasy.*;

PeasyCam cam; // Camera
ControlP5 cp5; // User Interface 
Table table, deweyTable;
ArrayList<String> deweyNames;
int numRows, numCols;
float maxRate;

// store every Checkout
ArrayList<Checkout> allCheckouts = new ArrayList<Checkout>();
ArrayList<Checkout> checkouts = new ArrayList<Checkout>();
ArrayList<Checkout> running = new ArrayList<Checkout>();
HashMap<Integer, Boolean> isInUse = new HashMap<Integer, Boolean>();
Boolean afterInitialized = false;

// ensure each deweyClass gets its own color
HashMap<Integer, Integer> colorMap = new HashMap<Integer, Integer>();

// allow for Checkouts to be sorted based on checkout day
public class CheckoutsComparator implements Comparator<Checkout> {
  @Override public int compare(Checkout c1, Checkout c2) {
    return c2.day - c1.day;
  }
}

void setup() {
  fullScreen(P3D);
  smooth();
  frameRate(60);
  cam = new PeasyCam(this, 200);
  cam.rotateX(170);
  // change sensitivity of built-in mouse wheel zoom
  cam.setWheelScale(.6); // 1.0 by default
  cam.setDistance(3600);
  cp5 = new ControlP5(this);

  // load dewey classes table
  deweyTable = loadTable("deweyClasses.csv");
  deweyNames = new ArrayList<String>();
  for (int i=0; i<deweyTable.getRowCount(); i++) {
    if (deweyTable.getInt(i, 0) % 100 == 0 ) {
      deweyNames.add(deweyTable.getString(i, 1));
    }
  }

  // load checkouts table
  table = loadTable("12month.csv", "header");
  numRows = table.getRowCount();
  println("numRows: " + numRows);

  // process each row of data. append each book to list
  for (int i=0; i<numRows; i++) {
    Checkout cout = new Checkout(table.getString(i, 0), table.getString(i, 1), 
      table.getFloat(i, 3), table.getString(i, 2), table.getString(i, 4));

    maxRate = max(maxRate, cout.rate);

    // if the colorMap does not have a color for this dewey, insert a random one for the color
    if (!colorMap.containsKey(cout.dewey)) {
      colorMap.put(cout.dewey, color(random(255), random(255), random(255)));
    }   
    allCheckouts.add(cout);
  }

  // sort checkouts by day
  Collections.sort(allCheckouts, new CheckoutsComparator());
  //checkouts = allCheckouts;

  // set up GUI for user Dewey input
  Textfield Dewey = cp5.addTextfield("dewey")
    .setPosition(20, 30)
    .setSize(70, 20)
    .setFocus(true)
    .setColor(color(0))
    .setColorBackground(color(255))
    .setColorActive(color(255))
    .setColorForeground(color(255))
    .setCaptionLabel("Please Enter A Dewey Number from 0 to 990");

  Dewey.getCaptionLabel().getStyle().marginTop  = -16; 
  Dewey.getCaptionLabel().getStyle().marginLeft  = 85;
  cp5.setAutoDraw(false);

  for (int i=0; i<=900; i+=100) {
    dewey(String.valueOf(i));
  }
  afterInitialized = true;
}

int curDay = 0;
String curDate = "";
int timeGap = 0;
void draw() {
  background(0);
  lights();

  // list the dewey classes currently in display
  int textOffset = 75;
  for (int dewey : isInUse.keySet()) {
    fill(colorMap.get(dewey));
    //textSize(30);  
    cam.beginHUD();  //Stopping peasy cam
    textFont(createFont("Arial", 15));
    textAlign(LEFT);
    text(dewey + deweyNames.get(dewey/100), 20, textOffset);
    textOffset+=15;
    cam.endHUD();
  }

  // remove those that passed boundary
  ArrayList<Checkout> toRemove = new ArrayList<Checkout>();

  // move from checkouts list to running every second
  if (!checkouts.isEmpty() && timeGap%10==0) {
    Checkout cout = checkouts.get(checkouts.size()-1);
    running.add(cout);
    checkouts.remove(cout);
  }


  // draw current running Checkouts
  for (int i=0; i<running.size(); i++) {
    Checkout cout = running.get(i);
    if (cout.day > curDay) {
      curDay = cout.day;
      curDate = cout.date;
    }
    fill(colorMap.get(cout.dewey));
    if (cout.itemType.equals("acbk")) {
      stroke(colorMap.get(cout.dewey));
    } else {
      noStroke();
    }
    cout.x+=map(cout.rate, maxRate, 1.0, .5, .1);

    if (cout.x > 2500) {
      toRemove.add(cout);
    } else {
      cout.draw(); 
      // draw date
      fill(255);
      textMode(SHAPE);
      textSize(150);      
      textAlign(CENTER);
      text(curDate, 0, -800);
      running.set(i, cout);
    }
  }

  noFill();
  stroke(150);
  strokeWeight(2);
  box(5000, 960, 960);

  // remove from running
  for (Checkout cout : toRemove) {
    running.remove(cout);
  }

  cam.beginHUD();  //Stopping peasy cam
  cp5.draw();
  cam.endHUD();

  timeGap++;
}

public void dewey(String theText) {
  // prevent dewey hashMap from resetting during setup and everytime a new string is entered
  if (afterInitialized) {
    isInUse.clear(); 
    afterInitialized = false;
  }

  // automatically receives results from controller input
  println("a textfield event for controller 'input' : "+theText);
  int selectedDewey = int(theText);
  selectedDewey -= selectedDewey%100;
  if (!isInUse.containsKey(selectedDewey) && (selectedDewey <= 900 || selectedDewey >= 0)) {
    isInUse.put(selectedDewey, true);
    checkouts.clear();
    for (Checkout cout : allCheckouts) {
      if (isInUse.keySet().contains(cout.dewey)) {
        checkouts.add(cout);
      }
    }
    Collections.sort(checkouts, new CheckoutsComparator());
    curDay = 0;
    curDate = "";
  }
}
